@@ -19,6 +19,14 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
+#
+# IMPLEMENTATION NOTE
+# Since Xapian is not thread safe, but single-writer-multiple-reader safe,
+# we must only do write operations from our worker thread, never from the
+# main thread (or as callbacks to in a gmainloop). Note that close()
+# and friends also may cause writes to the db.
+#
+
 #
 # TODO
 #
@@ -42,6 +50,7 @@ from Queue import Queue, Empty
 import threading
 from urllib import quote as url_escape, unquote as url_unescape
 import gobject, gio
+from cStringIO import StringIO
 
 from zeitgeist.datamodel import Symbol, StorageState, ResultType, TimeRange, NULL_EVENT, NEGATION_OPERATOR
 from _zeitgeist.engine.datamodel import Event, Subject
@@ -53,6 +62,8 @@ logging.basicConfig(level=logging.DEBUG)
 log = logging.getLogger("zeitgeist.fts")
 
 INDEX_FILE = os.path.join(constants.DATA_PATH, "fts.index")
+INDEX_VERSION = "1"
+INDEX_LOCK = threading.Lock()
 FTS_DBUS_OBJECT_PATH = "/org/gnome/zeitgeist/index/activity"
 FTS_DBUS_INTERFACE = "org.gnome.zeitgeist.Index"
 
@@ -84,6 +95,23 @@ COALESCING_RESULT_TYPES = [ \
 	ResultType.LeastPopularActor,
 ]
 
+# Xapian has a maximum term length of 245 bytes and Bad Things(TM) happen
+# if you bust that. We use the cap_string() function to control this.
+MAX_TERM_LENGTH = 245
+
+def synchronized(lock):
+    """ Synchronization decorator. """
+
+    def wrap(f):
+        def newFunction(*args, **kw):
+            lock.acquire()
+            try:
+                return f(*args, **kw)
+            finally:
+                lock.release()
+        return newFunction
+    return wrap
+
 class Deletion:
 	"""
 	A marker class that marks an event id for deletion
@@ -91,6 +119,20 @@ class Deletion:
 	def __init__ (self, event_id):
 		self.event_id = event_id
 
+class Reindex:
+	"""
+	Marker class that tells the worker thread to rebuild the entire index.
+	On construction time all events are pulled out of the zg_engine
+	argument and stored for later processing in the worker thread.
+	This avoid concurrent access to the ZG sqlite db from the worker thread.
+	"""
+	def __init__ (self, zg_engine):
+		all_events = zg_engine.find_events(TimeRange.always(),
+		                                   [], StorageState.Any,
+		                                   sys.maxint,
+		                                   ResultType.MostRecentEvents)
+		self.all_events = all_events
+
 class SearchEngineExtension (Extension, dbus.service.Object):
 	"""
 	Full text indexing and searching extension for Zeitgeist
@@ -103,8 +145,7 @@ class SearchEngineExtension (Extension, dbus.service.Object):
 		                             FTS_DBUS_OBJECT_PATH)
 		self._indexer = Indexer(self.engine)
 	
-	def pre_insert_event(self, event, sender):
-		# Fix when Zeitgeist 0.5.1 hits the street use post_insert_event() instead
+	def post_insert_event(self, event, sender):
 		self._indexer.index_event (event)
 		return event
 
@@ -127,6 +168,18 @@ class SearchEngineExtension (Extension, dbus.service.Object):
 		                                         offset, count, result_type)
 		return self._make_events_sendable (events), hit_count
 	
+	@dbus.service.method(FTS_DBUS_INTERFACE,
+	                     in_signature="",
+	                     out_signature="")
+	def ForceReindex(self):
+		"""
+		DBus method to force a reindex of the entire Zeitgeist log.
+		This method is only intended for debugging purposes and is not
+		considered blessed public API.
+		"""
+		log.debug ("Received ForceReindex request over DBus.")
+		self._indexer._queue.put (Reindex (self.engine))
+	
 	def _make_events_sendable(self, events):
 		for event in events:
 			if event is not None:
@@ -150,6 +203,31 @@ def mangle_uri (uri):
 			result += c
 	return result
 
+def cap_string (s, nbytes=MAX_TERM_LENGTH):
+	"""
+	If s has more than nbytes bytes (not characters) then cap it off
+	after nbytes bytes in a way still producing a valid utf-8 string.
+	
+	Assumes that s is a utf-8 string.
+	
+	This function useful for working with Xapian terms because Xapian has
+	a max term length of 245 (which is not very well documented, but see
+	http://xapian.org/docs/omega/termprefixes.html).
+	"""
+	# Check if we can fast-path this string
+	if (len(s.encode("utf-8")) <= nbytes):
+		return s
+	
+	# We use a StringIO here to avoid mem thrashing via naiive
+	# string concatenation. See fx. http://www.skymind.com/~ocrow/python_string/
+	buf = StringIO()
+	for char in s :
+		if buf.tell() >= nbytes - 1 :
+			return buf.getvalue()
+		buf.write(char.encode("utf-8"))
+	
+	return unicode(buf.getvalue().decode("utf-8"))
+
 def expand_type (type_prefix, uri):
 	"""
 	Return a string with a Xapian query matching all child types of 'uri'
@@ -176,9 +254,19 @@ class Indexer:
 	
 	def __init__ (self, engine):
 		self._engine = engine
+
+		# Activate support for the CJK tokenizer
+		os.environ["XAPIAN_CJK_NGRAM"] = "1"
 	
 		log.debug("Opening full text index: %s" % INDEX_FILE)
-		self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OPEN)
+		try:
+			self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OPEN)
+		except xapian.DatabaseError, e:
+			# This is a slightly broad catch (one would expect only
+			# a DatabaseCorruptError) but Xapian seems to be throwing
+			# a generic DatabaseError in a lot of cases
+			log.warn("Full text index corrupted: '%s'. Rebuilding index." % e)
+			self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OVERWRITE)
 		self._tokenizer = indexer = xapian.TermGenerator()
 		self._query_parser = xapian.QueryParser()
 		self._query_parser.set_database (self._index)
@@ -202,7 +290,7 @@ class Indexer:
 		      xapian.NumberValueRangeProcessor(VALUE_TIMESTAMP, "ms", False))
 		self._query_parser.set_default_op(xapian.Query.OP_AND)
 		self._enquire = xapian.Enquire(self._index)
-		
+
 		# Cache of parsed DesktopEntrys
 		self._desktops = {}
 		
@@ -212,43 +300,42 @@ class Indexer:
 		self._worker = threading.Thread(target=self._worker_thread,
 		                                name="IndexWorker")
 		self._worker.daemon = True
-		self._worker.start()
 		
-		self._check_index ()
+		# We need to defer the index checking until after ZG has completed
+		# full setup. Hence the idle handler.
+		# We also don't start the worker until after we've checked the index
+		gobject.idle_add (self._check_index_and_start_worker)
+	
+	@synchronized (INDEX_LOCK)
+	def _check_index_and_start_worker (self):
+		"""
+		Check whether we need a rebuild of the index.
+		Returns True if the index is good. False if a reindexing has
+		been commenced.
 		
-	def _check_index (self):
-		if self._index.get_doccount() == 0:
+		This method should be called from the main thread and only once.
+		It starts the worker thread as a side effect.
+
+		We are clearing the queue, because there may be a race when an
+		event insertion / deletion is already queued and our index
+		is corrupted. Creating a new queue instance should be safe,
+		because we're running in main thread as are the index_event
+		and delete_event methods, and the worker thread wasn't yet
+		started.
+		"""
+		if self._index.get_metadata("fts_index_version") != INDEX_VERSION:
+			log.info("Index must be upgraded. Doing full rebuild")
+			self._queue = Queue(0)
+			self._queue.put(Reindex(self._engine))
+		elif self._index.get_doccount() == 0:
 			# If the index is empty we trigger a rebuild
 			# We must delay reindexing until after the engine is done setting up
 			log.info("Empty index detected. Doing full rebuild")
-			gobject.idle_add (self._reindex)
-		else:
-			# If the index doesn't use the zgsu prefix, it must be old-style,
-			# and we must rebuild it
-			query = self._query_parser.parse_query ("zgsu:file*",
-		                                            self.QUERY_PARSER_FLAGS)
-			self._enquire.set_query (query)
-			hits = self._enquire.get_mset (0, 1)
-			hit_count = hits.get_matches_estimated()
-			if hit_count == 0:
-				log.info ("Old index format detected. Rebuilding index.")
-				gobject.idle_add (self._reindex)
-	
-	def _reindex (self):
-		"""
-		Index everything in the ZG log
-		"""
-		self._index.close ()
-		self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OVERWRITE)
-		self._query_parser.set_database (self._index)
-		self._enquire = xapian.Enquire(self._index)
+			self._queue = Queue(0)
+			self._queue.put(Reindex(self._engine))
 		
-		all_events = self._engine.find_events(TimeRange.always(),
-		                                      [], StorageState.Any,
-		                                      sys.maxint,
-		                                      ResultType.MostRecentEvents)
-		log.info("Preparing to index %s events" % len(all_events))
-		for e in all_events : self._queue.put(e)
+		# Now that we've checked the index from the main thread we can start the worker
+		self._worker.start()
 	
 	def index_event (self, event):
 		"""
@@ -267,6 +354,7 @@ class Indexer:
 		self._queue.put (Deletion(event_id))
 		return		
 	
+	@synchronized (INDEX_LOCK)
 	def search (self, query_string, time_range=None, filters=None, offset=0, maxhits=10, result_type=100):
 		"""
 		Do a full text search over the indexed corpus. The `result_type`
@@ -354,6 +442,8 @@ class Indexer:
 				
 				if isinstance (event, Deletion):
 					self._delete_event_real (event.event_id)
+				elif isinstance (event, Reindex):
+					self._reindex (event.all_events)
 				else:
 					self._index_event_real (event)
 				
@@ -367,6 +457,25 @@ class Indexer:
 				else:
 					log.debug("No changes to index. Sleeping")
 	
+	@synchronized (INDEX_LOCK)
+	def _reindex (self, event_list):
+		"""
+		Index everything in the ZG log. The argument must be a list
+		of events. Typically extracted by a Reindex instance.
+		Only call from worker thread as it writes to the db and Xapian
+		is *not* thread safe (only single-writer-multiple-reader).
+		"""
+		self._index.close ()
+		self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OVERWRITE)
+		self._query_parser.set_database (self._index)
+		self._enquire = xapian.Enquire(self._index)
+	
+		self._index.set_metadata("fts_index_version", INDEX_VERSION)
+		
+		log.info("Preparing to rebuild index with %s events" % len(event_list))
+		for e in event_list : self._queue.put(e)
+
+	@synchronized (INDEX_LOCK)
 	def _delete_event_real (self, event_id):
 		"""
 		Look up the doc id given an event id and remove the xapian.Document
@@ -480,7 +589,7 @@ class Indexer:
 			
 				doc = self._tokenizer.get_document()
 				for cat in desktop.getCategories():
-					doc.add_boolean_term(FILTER_PREFIX_XDG_CATEGORY+cat.lower())
+					doc.add_boolean_term(cap_string(FILTER_PREFIX_XDG_CATEGORY+cat.lower()))
 		else:
 			log.debug("Unable to look up app info for %s" % actor)
 		
@@ -493,7 +602,7 @@ class Indexer:
 		# usually web URIs, are indexed in another way because there may
 		# be domain name etc. in there we want to rank differently
 		scheme, host, path = self._split_uri (url_unescape (uri))
-		if scheme == "file://" or not scheme:
+		if scheme == "file" or not scheme:
 			path, name = os.path.split(path)
 			self._tokenizer.index_text(name, 5)
 			self._tokenizer.index_text(name, 5, "N")
@@ -503,16 +612,21 @@ class Indexer:
 			while path and name:
 				weight = weight / 1.5
 				path, name = os.path.split(path)
-				self._tokenizer.index_text(name, weight)
+				self._tokenizer.index_text(name, int(weight))
 			
-		elif scheme == "mailto:":
+		elif scheme == "mailto":
 			tokens = host.split("@")
 			name = tokens[0]
 			self._tokenizer.index_text(name, 6)
 			if len(tokens) > 1:
 				self._tokenizer.index_text(" ".join[1:], 1)
 		else:
+			# We're cautious about indexing the path components of
+			# non-file URIs as some websites practice *extremely* long
+			# and useless URLs
 			path, name = os.path.split(path)
+			if len(name) > 30 : name = name[:30]
+			if len(path) > 30 : path = path[30]
 			if name:
 				self._tokenizer.index_text(name, 5)
 				self._tokenizer.index_text(name, 5, "N")
@@ -560,31 +674,33 @@ class Indexer:
 		"""Adds the filtering rules to the doc. Filtering rules will
 		   not affect the relevancy ranking of the event/doc"""
 		if event.interpretation:
-			doc.add_boolean_term (FILTER_PREFIX_EVENT_INTERPRETATION+event.interpretation)
+			doc.add_boolean_term (cap_string(FILTER_PREFIX_EVENT_INTERPRETATION+event.interpretation))
 		if event.manifestation:
-			doc.add_boolean_term (FILTER_PREFIX_EVENT_MANIFESTATION+event.manifestation)
+			doc.add_boolean_term (cap_string(FILTER_PREFIX_EVENT_MANIFESTATION+event.manifestation))
 		if event.actor:
-			doc.add_boolean_term (FILTER_PREFIX_ACTOR+mangle_uri(event.actor))
+			doc.add_boolean_term (cap_string(FILTER_PREFIX_ACTOR+mangle_uri(event.actor)))
 		
 		for su in event.subjects:
 			if su.uri:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_URI+mangle_uri(su.uri))
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_URI+mangle_uri(su.uri)))
 			if su.interpretation:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_INTERPRETATION+su.interpretation)
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_INTERPRETATION+su.interpretation))
 			if su.manifestation:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_MANIFESTATION+su.manifestation)
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_MANIFESTATION+su.manifestation))
 			if su.origin:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_ORIGIN+mangle_uri(su.origin))
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_ORIGIN+mangle_uri(su.origin)))
 			if su.mimetype:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_MIMETYPE+su.mimetype)
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_MIMETYPE+su.mimetype))
 			if su.storage:
-				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_STORAGE+su.storage)
+				doc.add_boolean_term (cap_string(FILTER_PREFIX_SUBJECT_STORAGE+su.storage))
 	
+	@synchronized (INDEX_LOCK)
 	def _index_event_real (self, event):
 		if not isinstance (event, Event):
 			log.error("Not an Event, found: %s" % type(event))
+			return
 		if not event.id:
-			log.debug("Not indexing event. Event has no id")
+			log.warning("Not indexing event. Event has no id")
 			return
 		
 		try:
@@ -599,8 +715,17 @@ class Indexer:
 		
 			for subject in event.subjects:
 				if not subject.uri : continue
+				
+				# By spec URIs can have arbitrary length. In reality that's just silly.
+				# The general online "rule" is to keep URLs less than 2k so we just
+				# choose to enforce that. We discard the entire event and not just the
+				# subject to avoid special casing this everywhere we handle events
+				if len(subject.uri) > 2000:
+					log.info ("URI too long (%s). Discarding: %s..."% (len(subject.uri), subject.uri[:30]))
+					return
 				log.debug("Indexing '%s'" % subject.uri)
 				self._index_uri (subject.uri)
+
 				self._index_text (subject.text)
 				
 				# If the subject URI is an actor, we index the .desktop also
@@ -666,6 +791,3 @@ class Indexer:
 		
 		return "%s..%sms" % (time_range.begin, time_range.end)
 
-if __name__ == "__main__":
-	indexer = Indexer(None)
-	print indexer._compile_filter_query([Event.new_for_values(subject_interpretation="http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#Document")])
